#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# ARCLI sample runner.
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging.config
import sys
import time
from aacommons.contentprovider.ContentProvider import FileContentProvider
from aacommons.dataparser.cli.CliParser import CliParserFactory
from aacommons.dataparser.cli.arcli.ArCliParserRequest import ArCliParserRequest


LOGGING_CONFIG = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)-.1s <%(module)s> %(funcName)s | %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'WARNING',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout', # Default is stderr
        },
        'debug': {
            'level': 'DEBUG',
            'formatter': 'standard',
            'class': 'logging.handlers.RotatingFileHandler',
            'filename': "arcli-parser.log",
            'maxBytes': 10 * 1024 * 1024,
            'backupCount': 10
        }
    },
    'loggers': {
        'aacommons': { 
            'handlers': ['debug'],
            'level': 'DEBUG',
            'propagate': False
        }
    } 
}
logging.config.dictConfig(LOGGING_CONFIG)


def printResultDataRecord(record):
    n = list(record.keys())[0]
    d = record[n]
    for dr in d:
        (doc, value) = dr
        print(n, doc, value)


def printResultDataRecords(records, f):
    for record in records:
        f(record)


def printResultData(rd):
    if 'regexes_entire' in rd:
        printResultDataRecords(rd['regexes_entire'], printResultDataRecord)
    if 'tables' in rd:
        printResultDataRecords(rd['tables'], printResultDataRecord)
    if 'regexes_byline' in rd:
        printResultDataRecords(rd['regexes_byline'], printResultDataRecord)


def printResult(r, indent):
    (command, resultData) = r
    if isinstance(resultData, dict) or resultData is None:
        # Terminal node
        print(indent, command, "->", str(resultData)[0:60], "...", str(resultData)[-40:])
        if resultData is not None:
            printResultData(resultData)
    elif isinstance(resultData, list):
        print(indent, command)
        for rd in resultData:
            printResult(rd, indent + "  ")
    else:
        raise Exception("not a tuple or list")


def main():
    # Instantiate parser
    contentProvider = FileContentProvider(sys.argv[1])
    cliParser = CliParserFactory.getParser("ARCLI", arCliConfig=contentProvider)

    # Load content
    f = open(sys.argv[4], 'r')
    content = f.read()
    f.close()

    # Parse
    parserRequest = ArCliParserRequest()
    parserRequest['Source'] = "me"
    parserRequest['Command'] = sys.argv[3]
    parserRequest['Stdout'] = content
    parserRequest['Label'] = sys.argv[2]
    parserRequest["LocalBeginTime"] = time.time()
    r = cliParser.parse(parserRequest)

    # Print results
    if r is None:
        print("NO RESULT")
    elif isinstance(r, tuple):
        printResult(r, "")
    else:
        print("UNEXPECTED RESULT")

if __name__ == '__main__':
    if len(sys.argv) != 5:
        print("%s: <arcli.json> <label> <command> <content>" % (sys.argv[0]))
        sys.exit(1)
    main()



#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# ARCLI parser.
#
# Authors: Thomas Bastian, Jeff Goff
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
import re
import time
from ..CliParserBase import CliParser
from .ArCliResult import ArCliResult
from aacommons.dataparser.cli.TableHelpers import contentAsListOfDict, contentAsParameterValueList


# Default field name separator
ND_FIELD_NAME_SEPARATOR='.'
# Internal, not exposed
TAG_ND_ID='nd.id'


# Logger
log = logging.getLogger(__name__)
logDisp = logging.getLogger(__name__ + "#dispatch")
logAr = logging.getLogger(__name__ + "#parser")


def arCliProcessRecordBeginEnd(record, source):
    # Pre-process source
    r_begin = None
    if 'begin' in record:
        recordBeginCompiledRegex, recordBeginRegex = record['begin']
        m = re.search(recordBeginCompiledRegex, source)
        # If a regex is specified it needs to match
        if m is None:
            logAr.debug("record.begin regex: [%s] did NOT match" % (recordBeginRegex))
            return None
        r_begin = m.start()
    r_end = None
    if 'end' in record:
        recordEndCompiledRegex, recordEndRegex = record['end']
        m = re.search(recordEndCompiledRegex, source)
        # If a regex is specified it needs to match
        if m is None:
            logAr.debug("record.end regex: [%s] did NOT match" % (recordEndRegex))
            return None
        r_end = m.start()
    if r_begin is not None or r_end is not None:
        if r_begin is None:
            # Default to beginning of CLI output
            r_begin = 0
        olen = len(source)
        if r_end is None:
            # Default to end of CLI output
            r_end = olen
        logAr.debug("record text [%d:%d] out of %d" % (r_begin, r_end, olen))
        return source[r_begin:r_end]
    return source


def arCliProcessRecords(records, command, cliOutput, P):
    # Walk through all record directives
    logAr.debug('records: %d, command: %s' % (len(records), command))
    all_records = []
    for record in records:
        logAr.debug("starting process record entry - %s", record)
        r_source = arCliProcessRecordBeginEnd(record, cliOutput)
        if r_source is None:
            logAr.debug("begin or end regex not matching, skipping record entry - %s" % record)
            continue
        r_options = record.get('roptions', {})
        (startCompiledRegex, startRegex, startRegexMode, startRegexOp) = record['start']
        hasStop = 'stop' in record
        if hasStop:
            (stopCompiledRegex, stopRegex, stopRegexMode, stopRegexOp) = record['stop']
        r_command = None
        r_start = None
        r_stop = None
        r_offset = 0
        r_regex_op = 'start'
        r_records = []
        while True:
            # Look for startRegex or stopRegex starting at current offset
            r_text = r_source[r_offset:]
            logAr.debug("record entry %s, regex op %s" % (record, r_regex_op))
            if r_regex_op == 'start':
                m = startRegexOp(startCompiledRegex, r_text)
                if hasStop:
                    # Next regex is 'stop'
                    r_regex_op = 'stop'
                    r_start = r_stop = r_command = None
            else:
                m = stopRegexOp(stopCompiledRegex, r_text)
                # Always fallback to start
                r_regex_op = 'start'
            if m is None:
                # TODO: check for stop case here
                logAr.debug("ending process record entry - %s", record)
                if r_start is not None:
                    # Last record
                    r_records.append((r_command, r_source[r_start:]))
                break
            if r_start is None:
                # Record start
                r_start = r_offset + m.end()
            else:
                # Record stop
                r_stop = r_offset + m.start()
                r_records.append((r_command, r_source[r_start:r_stop]))
                r_start = r_offset + m.end()
            # By default we merge original command with new derived command
            r_prefix = command + ":"
            if not r_options.get('prefix', True):
                r_prefix = ""
            r_command = r_prefix + r_text[m.start():m.end()].rstrip()
            r_offset += m.end()
        logAr.debug("record entry %s produced %d records" % (record, len(r_records)))
        all_records.extend(r_records)
    return all_records


def arCliProcessRegexes(regexes, cliOutput, filterMode, P):
    matched = 0
    for (compiledRegex, regex, regexMode, regexOp) in regexes:
        if regexMode != filterMode:
            continue
        m = regexOp(compiledRegex, cliOutput)
        if not m:
            log.debug("failed regex: [%s]" % regex)
            if logAr.isEnabledFor(logging.DEBUG):
                logAr.debug("failed output: [%s]" % cliOutput)
            continue

        if logAr.isEnabledFor(logging.DEBUG):
            logAr.debug("matched regex: [%s] -> %s" % (regex, m.groupdict()))

        P.update(m.groupdict())
        matched += 1
    return matched


def arCliParseTables(tables, cliOutput, T):
    for table in tables:
        name = table['name']
        kind = table['type']
        marker = table['marker']
        ttext = cliOutput
        tstart = None
        (tbegin_re, tbegin_regex) = table['begin']
        if tbegin_re != None:
            m = re.search(tbegin_re, cliOutput)
            # If a regex is specified it needs to match
            if m is None:
                logAr.debug("table begin regex %s not matching, skipping table" % tbegin_regex)
                continue
            tstart = m.start()
        tend = None
        (tend_re, tend_regex) = table['end']
        if tend_re != None:
            m = re.search(tend_re, cliOutput)
            # If a regex is specified it needs to match
            if m is None:
                logAr.debug("table end regex %s not matching, skipping table" % tend_regex)
                continue
            tend = m.start()
        if tstart is not None or tend is not None:
            if tstart is None:
                # Default to beginning of CLI output
                tstart = 0
            olen = len(cliOutput)
            if tend is None:
                # Default to end of CLI output
                tend = olen
            logAr.debug("table text [%d:%d] out of %d" % (tstart, tend, olen))
            ttext = cliOutput[tstart:tend]
        logAr.debug('parsing table: %s/%s/%s' % (name, kind, marker))
        if kind == 0:
            # Table
            tableOptions=table['toptions']
            ttranspose = False if tableOptions is None else tableOptions.get('nd.transpose', False)
            if ttranspose:
                # Do transposition
                headerList = []
                rows = contentAsListOfDict(ttext, marker=marker, headerList=headerList, tableOptions=tableOptions)
                logAr.debug("parsed table(transposition): %s, rows: %d" % (name, len(rows)))
                if len(rows) > 0 and len(headerList) > 0 and len(headerList[0]) >= 2:
                    # Transposition only makes sense with two or more columns
                    nrows = []
                    for k in headerList[0][1:]:
                        nrow = {}
                        nrow[headerList[0][0]] = k
                        for r in rows:
                            nrow[r[headerList[0][0]]] = r[k]
                        nrows.append(nrow)
                    rows = nrows
                    logAr.debug("parsed table(transposed): %s, rows: %d" % (name, len(rows)))
            else:
                rows = contentAsListOfDict(ttext, marker=marker, tableOptions=tableOptions)
                logAr.debug("parsed table: %s, rows: %d" % (name, len(rows)))
        else:
            # KVP
            _rows = contentAsParameterValueList(ttext, marker=marker)
            rows = [ _rows ]
            logAr.debug("parsed kvp: %s, columns: %d" % (name, len(rows)))
        T.append((table, rows))


def arCliParse(cliCommandDef, cliOutput, P, T, L):
    # Regexes
    regexes = cliCommandDef['regexes']
    if len(regexes) > 0:
        # Entire output against regexes that demand it
        arCliProcessRegexes(regexes, cliOutput, 0, P)

        # Line by line regexes need entire output broken down by lines
        if cliCommandDef['regexes_byline'] > 0:
            logAr.debug("regexes line by line %d" % (cliCommandDef['regexes_byline']))
            # TODO: should this be splitlines() instead?
            lines = cliOutput.split('\n')
            L.extend(lines)

    # Tables
    tables = cliCommandDef['tables']
    if len(tables) > 0:
        arCliParseTables(tables, cliOutput, T)


def _bodyNumberConversion(body, mappings):
    anomalies = { }
    for (k, v) in body.items():
        # Do nothing when v is None
        if v is None:
            anomalies[k] = (v, None)
            continue
        # Special processing for mappings
        if k in mappings:
            if mappings[k] == 'string':
                if type(v) == str:
                    # Already string
                    continue
                else:
                    # Convert to string
                    try:
                        body[k] = str(v)
                        continue
                    except Exception as e:
                        anomalies[k] = (v, e)
                        continue
            # Otherwise fall-through and convert to number

        # Try plain integer/long
        try:
            body[k] = int(v)
            continue
        except:
            # Swallow and fall-through
            pass
        # Try float
        try:
            body[k] = float(v)
            continue
        except ValueError:
            # Leave as is
            pass
        except Exception as e:
            # Report as failed conversion
            anomalies[k] = (v, e)
            continue
    return anomalies


def _arCliProcessField(body, field, ndContext, P=None, C=None):
    fieldName = field['name']
    when = field['when']
    if when != None:
        try:
            if not eval(when):
                logAr.debug('field: skipping %s: when is false' % (fieldName))
                return
        except:
            # When failed, treat as false
            logAr.debug('field: skipping %s: when failed' % (fieldName))
            return

    fieldEName = field['ename']
    if fieldEName != None:
        fieldName = eval(fieldEName)
        logAr.debug('field: ename: %s' % (fieldName))

    if 'P' in field and P != None and field['P'] in P:
        # P assignment
        logAr.debug('P: %s for field: %s' % (field['P'], fieldName))
        body[fieldName] = P[field['P']]
        return

    if 'C' in field and C != None and field['C'] in C:
        # C assignment
        logAr.debug('C: %s for field: %s' % (field['C'], fieldName))
        body[fieldName] = C[field['C']]
        return

    if 'eval' in field:
        # The expression argument is parsed and evaluated as a python expression
        # (technically speaking, a condition list) using the globals and locals
        # dictionaries as global and local namespace. If the globals dictionary
        # is present and lacks '__builtins__', the current globals are copied
        # into globals before expression is parsed. This means that expression
        # normally has full access to the standard __builtin__ module and
        # restricted environments are propagated. If the locals dictionary is
        # omitted it defaults to the globals dictionary. If both dictionaries
        # are omitted, the expression is executed in the environment where eval()
        # is called. The return value is the result of the evaluated expression.
        # eval(expression[, globals[, locals]])
        try:
            result = eval(field['eval'])
        except KeyError as e:
            logAr.warning("KeyError exception [%s] in field [%s], for eval [%s], needs when clause" % \
                (e, field['name'], field['eval']))
            if C is not None:
                logAr.warning("C: [%s]", C)
            if P is not None and len(P) > 0:
                logAr.warning("P: [%s]", P)
            return
        except IndexError as e:
            # protect against dodgy evals
            logAr.warning("IndexError exception [%s] in field [%s], for eval [%s], needs when clause" % \
                (e, field['name'], field['eval']))
            if C is not None:
                logAr.warning("C: [%s]", C)
            if P is not None and len(P) > 0:
                logAr.warning("P: [%s]", P)
            return

        body[fieldName] = result
        return


def _arCliProcessTableRegexes(source, tableDef, name, body, P, C):
    for tableRegex in tableDef['regexes']:
        tableColumn = tableRegex['column']
        if not tableColumn in C:
            continue
        tableWhen = tableRegex['when']
        if tableWhen != None:
            # Only process regex when evaluation returns true
            try:
                if not eval(tableWhen):
                    logAr.debug("[%s] skipping %s: when is false" % \
                                (source, tableColumn))
                    continue
            except:
                # When failed, treat as false
                logAr.debug("[%s] skipping %s: when failed" % \
                            (source, tableColumn))
                continue
        # tableRegex['regex'] is precompiled in NdContext to be a tuple, despite the .json
        # being a single regex string.
        (tableCompiledRe, tableRe, regexOp) = tableRegex['regex']
        m = regexOp(tableCompiledRe, C[tableColumn])
        if not m:
            logAr.debug("[%s] failed table regex: [%s] column:[%s] value:[%s]" % \
                        (source, tableRe, tableColumn, C[tableColumn]))
            continue
        tableDefaults = tableRegex['defaults']
        for (k, v) in m.groupdict().items():
            # handle defaults for certain cases where the groupdict
            # returns None intentionally, i.e. "HT mode" for legacy/non-HT
            if logAr.isEnabledFor(logging.DEBUG):
                logAr.debug("   [%s] [%s] td %s" % (k, v, tableDefaults))
            if v is None:
                newVal = None
                if tableDefaults is not None:
                    newVal = tableDefaults.get(k, None)
                if newVal is not None:
                    logAr.debug("[%s] [%s] [%s] update value [%s][%s] -> [%s]" % \
                               (source, tableColumn, C[tableColumn], k, v, newVal))
                    v = newVal
                else:
                    # No sense to store None value, skip it
                    logAr.debug("[%s] [%s] [%s] update value [%s][%s] -> no default value!" % \
                                (source, tableColumn, C[tableColumn], k, v))
                    continue
            body[name + ND_FIELD_NAME_SEPARATOR + k] = v
        # Needed in case fields reference P
        P.update(m.groupdict())


class ArCliParser(CliParser):
    def __init__(self, parserContext):
        super().__init__()
        self._parserContext = parserContext
        # Label is of the form <device class>/<qualifier>
        # For now we fill only by <device class>
        self._parserMessageHandlerByLabel = dict()
        self._parserMessageHandlerByLabel["ACX"] = ArCliMessageHandlerAcx("ACX/*")
        self._parserMessageHandlerByLabel["AOS"] = ArCliMessageHandlerAos("AOS/*")
        self._parserMessageHandlerByLabel["ASW"] = ArCliMessageHandlerAsw("ASW/*")
        self._parserMessageHandlerByLabel["COMWARE"] = ArCliMessageHandlerComware("COMWARE/*")
        self._parserMessageHandlerByLabel["IAP"] = ArCliMessageHandlerIap("IAP/*")

    def parse(self, parserData):
        # Lookup by full label
        label = parserData['Label']
        parserMessageHandler = self._parserMessageHandlerByLabel.get(label, None)
        if parserMessageHandler is None:
            # Lookup by device class
            dclass = label.split('/')[0]
            parserMessageHandler = self._parserMessageHandlerByLabel.get(dclass, None)
            if parserMessageHandler is None:
                raise Exception(label)
        results = parserMessageHandler.ARRESULT(None, parserData, self._parserContext)
        return results


####
#
# Base message handler, do not put product specific code in this class
#
class ArCliMessageHandler():
    def __init__(self, kind):
        self._byCommandHandler = self.getCommandHandlers()
        self._kind = kind
        self._log = log
        self._log.debug("ArCliMessageHandler inited - %s" % self._kind)

    def ARRESULT(self, topic, jsonData, ndContext):
        try:
            command = jsonData['Command']
            commandHandler = self._byCommandHandler.get(command, self.processArCli)
            if self._log.isEnabledFor(logging.DEBUG):
                self._log.debug("[%s] CLI cmd: [%s] hdl: %s" % \
                    (jsonData.get("Source", "unknown"), command, commandHandler.__name__))
            results = commandHandler(command, jsonData, ndContext)
            return results
        except:
            self._log.critical("Exception in label %s commandHandler %s " % (jsonData['Label'], commandHandler.__name__), exc_info=True)
            self._log.critical('-'*60)
            self._log.critical("-- topic: %s " % topic)
            self._log.critical("-- command: %s " % command)
            self._log.critical("-- JSON: [%s]" % jsonData)

    def dumpCommandHandlers(self):
        self._log.info("%s command handlers" % self._kind)
        for k, v in self._byCommandHandler.items():
            self._log.info("    %-24s  ->  %s" % (k, v.__name__))

    def getCommandHandlers(self):
        return None

    def processArCli(self, command, jsonData, ndContext):
        self._log.debug("[%s] [%s] [%s] - process" % \
            (jsonData['Source'], jsonData['Label'], command))
        if logAr.isEnabledFor(logging.DEBUG):
            logAr.debug(jsonData)

        ndStats = []
        t0 = time.perf_counter()
        P = dict() # Holds parameters by name
        arCliDocuments = ndContext.getArCliDocumentsByLabel(jsonData['Label'])
        if arCliDocuments is None:
            self._log.info("Label %s has no arCliDocuments" % jsonData['Label'])
            return
        cliCommandDef = arCliDocuments.get(command, None)

        if cliCommandDef is None:
            for (compiledRegex, _cliCommandDef) in arCliDocuments['_']:
                m = re.search(compiledRegex, command)
                if m != None:
                    cliCommandDef = _cliCommandDef
                    P.update(m.groupdict())
                    break
            if cliCommandDef is None:
                self._log.info("[%s] [%s] [%s] drop - no cliCommandDef" % \
                    (jsonData['Source'], jsonData['Label'], command))
                return

        # Records
        records = cliCommandDef['records']
        if len(records) > 0:
            # Preprocess CLI output for records
            precords = arCliProcessRecords(records, command, jsonData['Stdout'], P)
            self._log.info("extracted %d records from [%s]" % (len(precords), command))
            if logAr.isEnabledFor(logging.DEBUG):
                logAr.debug(precords)
            results = []
            for precord in precords:
                (r_command, r_stdout) = precord
                r_jsonData = jsonData.clone()
                r_jsonData['Stdout'] = r_stdout
                r_jsonData['Command'] = r_command
                r_jsonData['Status'] = len(r_stdout)
                # TODO: drop topic
                r_results = self.ARRESULT(None, r_jsonData, ndContext)
                if r_results is not None:
                    if isinstance(r_results, tuple):
                        results.append(r_results)
                    elif isinstance(r_results, list):
                        results.append((r_command, r_results))
                    else:
                        raise Exception("not a tuple or list")
                else:
                    results.append((r_command, None))
            return (command, results)

        source = jsonData['Source']
        localTime = jsonData['LocalBeginTime']
        stdout = jsonData['Stdout']
        outputTags = jsonData.getOutputTags(ndContext, source)
        bodyTags = jsonData.getBodyTags(ndContext, source)
        t1 = time.perf_counter()
        ndStats.append(('time', source, 'arcli.dispatch_ms', (t1 - t0) * 1000.0))

        result = None
        try:
            result = self.arCliHandler(cliCommandDef, ndContext, source, bodyTags, outputTags, localTime, stdout, P, ndStats, jsonData)
        except Exception as e:
            self._log.error("Exception in arCliHandler: %s" % e, exc_info=True)

        # Post ND statistics
        # TODO: rework
        if False and ndContext.isSystemFullStats():
            # Each entry has tuple (event_type, event_source, event_group, event_increment)
            for (event_type, event_source, event_group, event_increment) in ndStats:
                ndContext.addSystemFullStats(event_type, event_source, event_group, event_increment)
        return (command, result)

    def arCliHandler(self, cliCommandDef, ndContext, source, bodyTags, outputTags, timeStamp, cliOutput, P, ndStats, jsonData):
        t0 = time.perf_counter()
        # Add system and platform tags as parameters
        P['source'] = source
        P.update(bodyTags)
        # Holds tables
        T = []
        # Holds line by line
        L = []
        # Parse content
        t1 = time.perf_counter()
        arCliParse(cliCommandDef, cliOutput, P, T, L)
        t2 = time.perf_counter()

        if logAr.isEnabledFor(logging.DEBUG):
            logAr.debug('P: %s' % P)
            logAr.debug('len(P): %d' % len(P))
            logAr.debug('len(T): %d' % len(T))
            logAr.debug('len(L): %d' % len(L))

        t3 = time.perf_counter()
        ndStats.append(('time', source, 'arcli.prepare_ms', (t1 - t0) * 1000.0))
        ndStats.append(('time', source, 'arcli.parse_ms', (t2 - t1) * 1000.0))

        if not jsonData.processDocument(ndContext, source, cliCommandDef):
            self._log.debug("processDocument returned False, skip processing document")
            return None
        result = self.arCliProcessDocument(cliCommandDef, ndContext, source, bodyTags, outputTags, timeStamp,
                                           P, T, L, ndStats, jsonData)
        t3 = time.perf_counter()
        ndStats.append(('time', source, 'arcli.process_ms', (t3 - t2) * 1000.0))
        return result

    def arCliProcessDocument(self, cliCommandDef, ndContext, source, bodyTags, outputTags, timeStamp,
                             P, T, L, ndStats, jsonData):
        t0 = time.perf_counter()
        fields = cliCommandDef['fields']
        doc_type = cliCommandDef['document']
        result = ArCliResult()
        # Process regex
        result.addType(ArCliResult.TYPE_REGEXES_ENTIRE)
        if cliCommandDef['regexes_entire'] > 0:
            body = {}
            for field in fields:
                fieldName = field['name']
                # TODO: revisit filtering
                if False and not field['enabled']:
                    logAr.debug('field: skipping %s: not enabled' % (fieldName))
                    continue
                _arCliProcessField(body, field, ndContext, P)
            # Single document
            messages = result.addSection(ArCliResult.TYPE_REGEXES_ENTIRE)
            self.arCliMessagePost(ndContext, source, bodyTags, outputTags, timeStamp, doc_type, body, messages, jsonData)
        # regex process done
        t1 = time.perf_counter()
        # Process tables
        result.addType(ArCliResult.TYPE_TABLES)
        if len(T) > 0:
            savedP = P
            for (tableDef, rows) in T:
                name = tableDef['name']
                logAr.debug('processing tables rows: %s' % (name))
                messages = result.addSection(ArCliResult.TYPE_TABLES, name)
                for C in rows:
                    body = {}
                    P = dict()
                    P.update(savedP)
                    # Add row content as fields
                    for (k, v) in C.items():
                        body[name + ND_FIELD_NAME_SEPARATOR + k] = v
                    # Process table regexes
                    _arCliProcessTableRegexes(source, tableDef, name, body, P, C)
                    # Process fields
                    for field in fields:
                        fieldName = field['name']
                        # TODO: revisit filtering
                        if False and not field['enabled']:
                            logAr.debug('field: skipping %s: not enabled' % (fieldName))
                            continue
                        _arCliProcessField(body, field, ndContext, P, C)
                    # Single document per table row
                    # TODO: add a document override for the table and field
                    tableTimestamp = tableDef['timestamp']
                    if tableTimestamp != None:
                        # Evaluation returns floating point time, convert to ms
                        evalTableTimestamp = eval(tableTimestamp)
                        timeStampMillis = evalTableTimestamp * 1000.0
                        # Need to force a static id otherwise
                        idkey = eval(tableDef['idkey'])
                        body[TAG_ND_ID] = name + ("%f" % timeStampMillis) + idkey
                        self.arCliMessagePost(ndContext, source, bodyTags, outputTags, evalTableTimestamp, doc_type, body, messages, jsonData)
                    else:
                        self.arCliMessagePost(ndContext, source, bodyTags, outputTags, timeStamp, doc_type, body, messages, jsonData)
            P = savedP
        # T processing done
        t2 = time.perf_counter()
        # Process line by line regex
        result.addType(ArCliResult.TYPE_REGEXES_BYLINE)
        if len(L) > 0:
            regexes = cliCommandDef['regexes']
            logAr.debug('processing line by line regexes: %s' % (regexes))
            savedP = P
            for row in L:
                P = dict()
                arCliProcessRegexes(regexes, row, 1, P)
                if len(P) > 0:
                    messages = result.addSection(ArCliResult.TYPE_REGEXES_BYLINE)
                    body = {}
                    P.update(savedP)
                    for field in fields:
                        fieldName = field['name']
                        # TODO: revisit filtering
                        if False and not field['enabled']:
                            logAr.debug('field: skipping %s: not enabled' % (fieldName))
                            continue
                        _arCliProcessField(body, field, ndContext, P)
                    # Single document per line
                    self.arCliMessagePost(ndContext, source, bodyTags, outputTags, timeStamp, doc_type, body, messages, jsonData)
        # line by line done
        t3 = time.perf_counter()
        ndStats.append(('time', source, 'arcli.procdoc.proc_regex_ms', (t1 - t0) * 1000.0))
        ndStats.append(('time', source, 'arcli.procdoc.proc_T_ms', (t2 - t1) * 1000.0))
        ndStats.append(('time', source, 'arcli.procdoc.proc_L_ms', (t3 - t2) * 1000.0))
        return result.get()

    def arCliMessagePost(self, ndContext, source, bodyTags, outputTags, timeStamp, doc_type, body, messages, jsonData):
        if len(body) == 0:
            # No data was copied
            return
        # Add body tags
        body.update(bodyTags)
        # Body conversion
        anomalies = _bodyNumberConversion(body, ndContext.getArCliMappings())
        if len(anomalies) > 0:
            # Report conversion anomalies
            log.warning('number conversion: [%s] %s %s anomalies: %s' % (source, doc_type, body, anomalies))
        # Final object
        output = {}
        output['timestamp'] = float(timeStamp)
        if outputTags is not None:
            output.update(outputTags)
        output['body'] = body
        _id = body.get(TAG_ND_ID, None)
        if _id != None:
            del body[TAG_ND_ID]
            output[TAG_ND_ID] = _id
        outputMessages = jsonData.updateMessageTags(ndContext, source, doc_type, output)
        if outputMessages is not None:
            if self._log.isEnabledFor(logging.DEBUG):
                for (d, o) in outputMessages:
                    self._log.debug('[%s] %s %s' % (source, d, o))
            messages.extend(outputMessages)


########################################
#
# Per Product ARCLI Handlers Start Here
# As these grow longer they should probably move to their
# own files to avoid clutter
#

#
# ArubaOS-CX Switch
#
class ArCliMessageHandlerAcx(ArCliMessageHandler):
    def __init__(self, kind):
        super().__init__(kind)
        self._byCommandHandler = self.getCommandHandlers()
        self._log = logging.getLogger(__name__ + "#handler#acx")
        self.dumpCommandHandlers()

    def getCommandHandlers(self):
        byCommandHandler = dict()
        return byCommandHandler


#
# ArubaOS-Switch (Procurve)
#
class ArCliMessageHandlerAsw(ArCliMessageHandler):
    def __init__(self, kind):
        super().__init__(kind)
        self._byCommandHandler = self.getCommandHandlers()
        self._log = logging.getLogger(__name__ + "#handler#asw")
        self.dumpCommandHandlers()

    def getCommandHandlers(self):
        byCommandHandler = dict()
        return byCommandHandler


#
# Aruba IAP
#
class ArCliMessageHandlerIap(ArCliMessageHandler):
    def __init__(self, kind):
        super().__init__(kind)
        self._byCommandHandler = self.getCommandHandlers()
        self._log = logging.getLogger(__name__ + "#handler#iap")
        self.dumpCommandHandlers()

    def getCommandHandlers(self):
        byCommandHandler = dict()
        return byCommandHandler


#
# Comware Switch
#
class ArCliMessageHandlerComware(ArCliMessageHandler):
    def __init__(self, kind):
        super().__init__(kind)
        self._byCommandHandler = self.getCommandHandlers()
        self._log = logging.getLogger(__name__ + "#handler#comware")
        self.dumpCommandHandlers()

    def getCommandHandlers(self):
        byCommandHandler = dict()
        byCommandHandler['display cpu'] = self.processDisplayCpu
        byCommandHandler['display mpls ldp peer verbose'] = self.processDisplayMplsLdpPeerVerbose
        byCommandHandler['display ospf statistics packet'] = self.processDisplayOspfStatisticsPacket
        return byCommandHandler

    def processDisplayCpu(self, command, jsonData, ndContext):
        self._log.debug("process for source %s" % jsonData['Source'])
        source = jsonData['Source']
        localTime = jsonData['LocalBeginTime']
        outputTags = jsonData.getOutputTags(ndContext, source)
        bodyTags = jsonData.getBodyTags(ndContext, source)
        stdout = jsonData['Stdout'].replace('\r', '')
        # Expect definition to be present in arcli.json
        arCliDocuments = ndContext.getArCliDocumentsByLabel(jsonData['Label'])
        if arCliDocuments is None:
            return
        result = ArCliResult()
        result.addType(ArCliResult.TYPE_REGEXES_BYLINE)
        messages = result.addMessages(ArCliResult.TYPE_REGEXES_BYLINE)
        cliCommandDef = arCliDocuments.get(command, None)
        regexes = cliCommandDef['regexes']
        lines = stdout.splitlines()
        ln = 0
        while ln < len(lines):
            line = lines[ln].strip()
            if line == "":
                ln += 1
                continue
            P = dict()
            if arCliProcessRegexes(regexes, line, 0, P):
                # Slot \d+ CPU \d+ CPU usage:
                if (ln + 3) < len(lines):
                    # Read in values
                    cpu5second = int(lines[ln + 1].strip().split('%')[0])
                    cpu1minute = int(lines[ln + 2].strip().split('%')[0])
                    cpu5minutes = int(lines[ln + 3].strip().split('%')[0])
                    doc_type = "COMWARE_CPU"
                    body = {
                        'comware.cpu.slot': P['slot'],
                        'comware.cpu.id': P['cpu'],
                        'comware.cpu.util5second': cpu5second,
                        'comware.cpu.util1minute': cpu1minute,
                        'comware.cpu.util5minutes': cpu5minutes
                    }
                    self.arCliMessagePost(ndContext, source, bodyTags, outputTags, localTime, doc_type, body, messages, jsonData)
                    ln += 3
                    continue
            ln += 1
        return (command, result.get())

    def processDisplayMplsLdpPeerVerbose(self, command, jsonData, ndContext):
        self._log.debug("process for source %s" % jsonData['Source'])
        source = jsonData['Source']
        localTime = jsonData['LocalBeginTime']
        outputTags = jsonData.getOutputTags(ndContext, source)
        bodyTags = jsonData.getBodyTags(ndContext, source)
        result = ArCliResult()
        result.addType(ArCliResult.TYPE_REGEXES_BYLINE)
        messages = result.addMessages(ArCliResult.TYPE_REGEXES_BYLINE)
        stdout = jsonData['Stdout'].replace('\r', '')
        lines = stdout.splitlines()
        prefix = "comware.mpls.ldp.peer.verbose."
        processingPeer = None
        patternSessionUptime = re.compile("(\\d+)\\:(\\d+)\\:(\\d+) .*")
        patternMsgs = re.compile("(\\d+)\\/(\\d+)")
        patternKA = re.compile("(\\d+)\\/(\\d+)")
        ln = 0
        while ln < len(lines):
            line = lines[ln].strip()
            if line == "":
                if processingPeer is not None:
                    self.arCliMessagePost(ndContext, source, bodyTags, outputTags, localTime, "COMWARE_MPLS_LDP_PEER_VERBOSE", processingPeer, messages, jsonData)
                processingPeer = None
                ln += 1
                continue
            if line.startswith("Peer LDP ID      :"):
                # Detected new peer
                processingPeer = dict()
                processingPeer[prefix + "PeerId"] = line.split(':', 1)[1].strip()
                ln += 1
                continue
            if processingPeer is not None:
                if line.startswith("Session State"):
                    _sessionState = line.split(':', 1)[1].strip().split(' ', 1)[0].strip()
                    processingPeer[prefix + "sessionState"] = _sessionState
                if line.startswith("Session Up Time"):
                    _sessionUptime = line.split(':', 1)[1].strip()
                    splits = re.match(patternSessionUptime, _sessionUptime)
                    processingPeer[prefix + "sessionUptime"] = _sessionUptime
                    processingPeer[prefix + "sessionUptime_"] = int(splits.group(1)) * 86400 + \
                        int(splits.group(2)) * 3600 + int(splits.group(3)) * 60
                if line.startswith("Msgs Sent/Rcvd"):
                    splits = re.match(patternMsgs, line.split(':', 1)[1].strip())
                    processingPeer[prefix + "MsgsSent"] = int(splits.group(1))
                    processingPeer[prefix + "MsgsRcvd"] = int(splits.group(2))
                if line.startswith("KA Sent/Rcvd"):
                    splits = re.match(patternKA, line.split(':', 1)[1].strip())
                    processingPeer[prefix + "KASent"] = int(splits.group(1))
                    processingPeer[prefix + "KARcvd"] = int(splits.group(2))
            ln += 1
        if processingPeer is not None:
            self.arCliMessagePost(ndContext, source, bodyTags, outputTags, localTime, "COMWARE_MPLS_LDP_PEER_VERBOSE", processingPeer, messages, jsonData)
        return (command, result.get())

    def processDisplayOspfStatisticsPacket(self, command, jsonData, ndContext):
        self._log.debug("process for source %s" % jsonData['Source'])
        source = jsonData['Source']
        localTime = jsonData['LocalBeginTime']
        outputTags = jsonData.getOutputTags(ndContext, source)
        bodyTags = jsonData.getBodyTags(ndContext, source)
        result = ArCliResult()
        result.addType(ArCliResult.TYPE_REGEXES_BYLINE)
        messagesGlobal = result.addMessages(ArCliResult.TYPE_REGEXES_BYLINE)
        messagesInterface = result.addMessages(ArCliResult.TYPE_REGEXES_BYLINE)
        stdout = jsonData['Stdout'].replace('\r', '')
        lines = stdout.splitlines()
        prefixGlobal = "comware.ospf.stats.packet.global."
        processingGlobal = None
        patternGlobalIO = re.compile("(\\d+) +(\\d+) +(\\d+) +(\\d+) +(\\d+) +(\\d+)")
        processingArea = None
        prefixInterface = "comware.ospf.stats.packet.interface."
        processingInterface = None
        patternInterfaceIO = re.compile("(\\d+) +(\\d+) +(\\d+) +(\\d+) +(\\d+)")
        ln = 0
        while ln < len(lines):
            line = lines[ln].strip()
            if line.startswith("Waiting to send packet count: "):
                # Detected global stats
                processingGlobal = dict()
                processingGlobal[prefixGlobal + "waiting"] = int(line.split(':', 1)[1].strip())
                ln += 1
                continue
            if processingGlobal is not None:
                if line.startswith("Input : "):
                    _globalInput = line.split(':', 1)[1].strip()
                    splits = re.match(patternGlobalIO, _globalInput)
                    processingGlobal[prefixGlobal + "input.Hello"] = int(splits.group(1))
                    processingGlobal[prefixGlobal + "input.DD"] = int(splits.group(2))
                    processingGlobal[prefixGlobal + "input.LSR"] = int(splits.group(3))
                    processingGlobal[prefixGlobal + "input.LSU"] = int(splits.group(4))
                    processingGlobal[prefixGlobal + "input.ACK"] = int(splits.group(5))
                    processingGlobal[prefixGlobal + "input.Total"] = int(splits.group(6))
                if line.startswith("Output: "):
                    _globalOutput = line.split(':', 1)[1].strip()
                    splits = re.match(patternGlobalIO, _globalOutput)
                    processingGlobal[prefixGlobal + "output.Hello"] = int(splits.group(1))
                    processingGlobal[prefixGlobal + "output.DD"] = int(splits.group(2))
                    processingGlobal[prefixGlobal + "output.LSR"] = int(splits.group(3))
                    processingGlobal[prefixGlobal + "output.LSU"] = int(splits.group(4))
                    processingGlobal[prefixGlobal + "output.ACK"] = int(splits.group(5))
                    processingGlobal[prefixGlobal + "output.Total"] = int(splits.group(6))
            if line.startswith("Area: "):
                # Detected per Area stats
                if processingGlobal is not None:
                    # Publish previous
                    self.arCliMessagePost(ndContext, source, bodyTags, outputTags, localTime, "COMWARE_OSPF_STATISTICS_PACKET_GLOBAL", processingGlobal, messagesGlobal, jsonData)
                    processingGlobal = None
                processingArea = line.split(':', 1)[1].strip()
                ln += 1
                continue
            if line.startswith("Interface: "):
                if processingInterface is not None:
                    # Publish previous
                    self.arCliMessagePost(ndContext, source, bodyTags, outputTags, localTime, "COMWARE_OSPF_STATISTICS_PACKET_INTERFACE", processingInterface, messagesInterface, jsonData)
                processingInterface = dict()
                processingInterface[prefixInterface + "area"] = processingArea
                processingInterface[prefixInterface + "interface"] = line.split(':', 1)[1].strip()
            if processingInterface is not None:
                if line.startswith("Input : "):
                    _interfaceInput = line.split(':', 1)[1].strip()
                    splits = re.match(patternInterfaceIO, _interfaceInput)
                    processingInterface[prefixInterface + "input.DD"] = int(splits.group(1))
                    processingInterface[prefixInterface + "input.LSR"] = int(splits.group(2))
                    processingInterface[prefixInterface + "input.LSU"] = int(splits.group(3))
                    processingInterface[prefixInterface + "input.ACK"] = int(splits.group(4))
                    processingInterface[prefixInterface + "input.Total"] = int(splits.group(5))
                if line.startswith("Output: "):
                    _interfaceOutput = line.split(':', 1)[1].strip()
                    splits = re.match(patternInterfaceIO, _interfaceOutput)
                    processingInterface[prefixInterface + "output.DD"] = int(splits.group(1))
                    processingInterface[prefixInterface + "output.LSR"] = int(splits.group(2))
                    processingInterface[prefixInterface + "output.LSU"] = int(splits.group(3))
                    processingInterface[prefixInterface + "output.ACK"] = int(splits.group(4))
                    processingInterface[prefixInterface + "output.Total"] = int(splits.group(5))
            ln += 1
        if processingInterface is not None:
            self.arCliMessagePost(ndContext, source, bodyTags, outputTags, localTime, "COMWARE_OSPF_STATISTICS_PACKET_INTERFACE", processingInterface, messagesInterface, jsonData)
        return (command, result.get())


#
# ArubaOS Controller
#
class ArCliMessageHandlerAos(ArCliMessageHandler):
    def __init__(self, kind):
        super().__init__(kind)
        self._byCommandHandler = self.getCommandHandlers()
        self._log = logging.getLogger(__name__ + "#handler#aos")
        self._log.debug("ArCliMessageHandlerAos inited")
        self.dumpCommandHandlers()

    def getCommandHandlers(self):
        byCommandHandler = dict()
        byCommandHandler['show datapath message-queue counters'] = self.processShowDatapathMessageQueueCounters
        return byCommandHandler

    def processShowDatapathMessageQueueCounters(self, command, jsonData, ndContext):
        self._log.debug("process for source %s" % jsonData['Source'])
        source = jsonData['Source']
        timeStamp = jsonData['LocalBeginTime']
        outputTags = jsonData.getOutputTags(ndContext, source)
        bodyTags = jsonData.getBodyTags(ndContext, source)
        result = ArCliResult()
        result.addType(ArCliResult.TYPE_REGEXES_BYLINE)
        content = jsonData['Stdout']
        sosmq_start = content.find("Datapath SOS Message Queue Statistics")
        dpimq_start = content.find("Datapath DPI Message Queue Statistics")
        if sosmq_start >= 0:
            # Process SOS section
            messagesSOS = result.addMessages(ArCliResult.TYPE_REGEXES_BYLINE)
            if dpimq_start >= 0:
                content_sos = content[sosmq_start:dpimq_start]
            else:
                content_sos = content[sosmq_start:]
            self.processShowDatapathMessageQueueCountersSos(ndContext, source, bodyTags, outputTags, timeStamp, content_sos, messagesSOS, jsonData)
        if dpimq_start >= 0:
            # Process DPI section
            messagesDPI = result.addMessages(ArCliResult.TYPE_REGEXES_BYLINE)
            content_dpi = content[dpimq_start:]
            self.processShowDatapathMessageQueueCountersDpi(ndContext, source, bodyTags, outputTags, timeStamp, content_dpi, messagesDPI, jsonData)
        return (command, result.get())

    def processShowDatapathMessageQueueCountersSos(self, ndContext, source, bodyTags, outputTags, timeStamp, content, messages, jsonData):
        cpus = None
        lines = content.splitlines()
        for line in lines:
            line = line.strip()
            if line.startswith("Cpu-->"):
                cpus = line.replace("Cpu-->", "").split()
                continue
            if line.startswith("Opcode"):
                continue
            if line.startswith("---"):
                continue
            if line.startswith("Datapath"):
                continue
            if line == "":
                cpus = None
                continue
            if cpus is None:
                continue
            step = 9
            opcode = line[0:step].strip()
            i = step
            for cpu in cpus:
                _hprio = line[i:i+step]
                if _hprio.strip() == "":
                    hprio = None
                else:
                    hprio = int(_hprio, 16)
                _lprio = line[i+step:i+2*step]
                if _lprio.strip() == "":
                    lprio = None
                else:
                    lprio = int(_lprio, 16)
                i += 2 * step
                if hprio is None and lprio is None:
                    continue
                body = {}
                body['aos.dp.message-queue.counters.sos.cpu'] = str(cpu)
                body['aos.dp.message-queue.counters.sos.opcode'] = str(opcode)
                if hprio is not None:
                    body['aos.dp.message-queue.counters.sos.highprio'] = hprio
                if lprio is not None:
                    body['aos.dp.message-queue.counters.sos.lowprio'] = lprio
                self.arCliMessagePost(ndContext, source, bodyTags, outputTags, timeStamp, "AOS_DP_MESSAGE-QUEUE_COUNTERS_SOS", body, messages, jsonData)

    def processShowDatapathMessageQueueCountersDpi(self, ndContext, source, bodyTags, outputTags, timeStamp, content, messages, jsonData):
        cpus = None
        lines = content.splitlines()
        for line in lines:
            line = line.strip()
            if line.startswith("Cpu->Opcode"):
                cpus = line.replace("Cpu->Opcode", "").split()
                continue
            if line.startswith("---"):
                continue
            if line.startswith("Datapath"):
                continue
            if line == "":
                cpus = None
                continue
            if cpus is None:
                continue
            step = 9
            opcode = line[0:15].strip()
            i = 20
            for cpu in cpus:
                _value = line[i:i+step]
                if _value.strip() == "":
                    value = None
                else:
                    value = int(_value, 16)
                i += step
                if value is None:
                    continue
                body = {}
                body['aos.dp.message-queue.counters.dpi.cpu'] = str(cpu)
                body['aos.dp.message-queue.counters.dpi.opcode'] = str(opcode)
                body['aos.dp.message-queue.counters.dpi.value'] = value
                self.arCliMessagePost(ndContext, source, bodyTags, outputTags, timeStamp, "AOS_DP_MESSAGE-QUEUE_COUNTERS_DPI", body, messages, jsonData)



#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# ARCLI documents.
#
# Authors: Thomas Bastian, Albert Pang
#
'''

'''
# Parameters
'''
'''
#
'''

import jstyleson
import logging
import re
from json import loads


# Logger
log = logging.getLogger(__name__)


class ArCliDocuments():
    def __init__(self, contentProvider, fieldNameSeparator):
        self.load(contentProvider, fieldNameSeparator)

    def load(self, contentProvider, fieldNameSeparator):
        log.info('loading: %s' % (contentProvider))
        arCliContent = contentProvider.getContent()
        arCliDefs = jstyleson.loads(arCliContent)
        arCliSystemMappings = arCliDefs['arcliSystemMappings']
        self.arCliMappings = arCliSystemMappings
        arCliDocumentsByLabel = dict()
        allDocumentDefs = arCliDefs['arcliDocuments']
        for k in allDocumentDefs.keys():
            label = allDocumentDefs[k]['label']
            isDefault = allDocumentDefs[k].get('default', False)
            extendLabel = allDocumentDefs[k].get('extend', None)
            log.info('loading: %s, label: %s, default: %s, extend: %s' %
                     (k, label, isDefault, extendLabel))
            if extendLabel is not None:
                # Find corresponding document key
                _extendK = None
                for extendK in allDocumentDefs.keys():
                    if allDocumentDefs[extendK]['label'] == extendLabel:
                        _extendK = extendK
                if _extendK is None:
                    raise Exception("extend label %s referenced by %s not found" % (extendLabel, k))
                # Merge documents, key "_" needs special treatment
                labeledDocumentDefs = self.load_(arCliDefs[_extendK], fieldNameSeparator)
                labeledDocumentDefsExtend = self.load_(arCliDefs[k], fieldNameSeparator)
                labeledDocumentDefsExtend_ = labeledDocumentDefsExtend.pop("_")
                labeledDocumentDefs.update(labeledDocumentDefsExtend)
                for (k, v) in labeledDocumentDefsExtend_:
                    i = 0
                    while i < len(labeledDocumentDefs["_"]):
                        (k1, _) = labeledDocumentDefs["_"][i]
                        if k == k1:
                            # Replace same key
                            labeledDocumentDefs["_"][i] = (k, v)
                            i = -1
                            break
                        i += 1
                    if i >= 0:
                        # Not found, append key
                        labeledDocumentDefs["_"].append((k, v))
            else:
                labeledDocumentDefs = self.load_(arCliDefs[k], fieldNameSeparator)
            arCliDocumentsByLabel[label] = labeledDocumentDefs
            if isDefault:
                arCliDocumentsByLabel['default/default'] = labeledDocumentDefs
        self.arCliDocumentsByLabel = arCliDocumentsByLabel

    def load_(self, arCliDocumentDefs, fieldNameSeparator):
        arCliDocuments = dict()
        arCliDocuments['_'] = []
        for (cliCommand, loadedCliCommandDef) in arCliDocumentDefs.items():
            log.debug('CLI command: %s' % cliCommand)
            cliCommandDef = dict()

            cliCommandDef['document'] = loadedCliCommandDef['document']
            log.debug('document: %s', cliCommandDef['document'])

            cliCommandParameters = loadedCliCommandDef.get('parameters', False)

            cliCommandGroup = 'none'
            cliCommandGroup = loadedCliCommandDef.get('group', cliCommandGroup)
            log.debug('default group: %s', cliCommandGroup)
            cliCommandDef['group'] = cliCommandGroup

            loadedCommandRecords = loadedCliCommandDef.get('records', None)
            log.debug('loaded records: %s', loadedCommandRecords)
            cliCommandRecords = []
            cliCommandDef['records'] = cliCommandRecords
            if loadedCommandRecords != None:
                for recordDef in loadedCommandRecords:
                    record = dict()
                    # begin and end (optional, default: None)
                    recordBegin = recordDef.get('begin', None)
                    if recordBegin is not None:
                        record['begin'] = (re.compile(recordBegin, flags=re.MULTILINE), recordBegin)
                    recordEnd = recordDef.get('end', None)
                    if recordEnd is not None:
                        record['end'] = (re.compile(recordEnd, flags=re.MULTILINE), recordEnd)
                    # start mandatory
                    record['start'] = parseArCliRegex(cliCommandDef, recordDef['start'])
                    # stop optional
                    if 'stop' in recordDef:
                        record['stop'] = parseArCliRegex(cliCommandDef, recordDef['stop'])
                    # roptions optional
                    recordOptions = recordDef.get('roptions', None)
                    if recordOptions is not None:
                        record['roptions'] = recordOptions
                    cliCommandRecords.append(record)
            log.debug('loaded %d records', len(cliCommandDef['records']))

            loadedCommandRegexes = loadedCliCommandDef.get('regexes', None)
            log.debug('loaded regexes: %s', loadedCommandRegexes)
            cliCommandRegexes = []
            cliCommandDef['regexes'] = cliCommandRegexes
            cliCommandDef['regexes_byline'] = 0
            if loadedCommandRegexes != None:
                for regexDef in loadedCommandRegexes:
                    cliCommandRegex = parseArCliRegex(cliCommandDef, regexDef)
                    cliCommandRegexes.append(cliCommandRegex)
            log.debug('loaded %d regexes (by line)', cliCommandDef['regexes_byline'])
            cliCommandDef['regexes_entire'] = \
                len(cliCommandRegexes) - cliCommandDef['regexes_byline']
            log.debug('loaded %d regexes (entire)', cliCommandDef['regexes_entire'])

            loadedCommandTables = loadedCliCommandDef.get('tables', None)
            log.debug('loaded tables: %s', loadedCommandTables)
            cliCommandTables = []
            cliCommandDef['tables'] = cliCommandTables
            if loadedCommandTables != None:
                for tableDef in loadedCommandTables:
                    cliCommandTable = parseArCliTable(cliCommandDef, tableDef,
                                                      fieldNameSeparator, self.arCliMappings)
                    cliCommandTables.append(cliCommandTable)

            loadedCommandFields = loadedCliCommandDef.get('fields', None)
            log.debug('loaded fields: %s', loadedCommandFields)
            cliCommandFields = []
            cliCommandDef['fields'] = cliCommandFields
            if loadedCommandFields != None:
                for fieldDef in loadedCommandFields:
                    log.debug('processing field: %s', fieldDef)
                    cliCommandField = parseArCliField(cliCommandDef, fieldDef,
                                                      self.arCliMappings)
                    cliCommandFields.append(cliCommandField)

            if cliCommandParameters:
                # Need regex
                log.debug("add regex command: %s" % (cliCommand))
                arCliDocuments['_'].append((re.compile(cliCommand), cliCommandDef))
            else:
                arCliDocuments[cliCommand] = cliCommandDef
        log.info('done')
        return arCliDocuments

    def getByLabel(self, label):
        if self.arCliDocumentsByLabel is None:
            log.debug("no documents available")
            return None
        arCliDocuments = self.arCliDocumentsByLabel.get(label, None)
        if arCliDocuments is None:
            log.debug("no documents for label [%s]" % label)
        return arCliDocuments

    def getMappings(self):
        return self.arCliMappings


def parseArCliRegex(cliCommandDef, regexDef):
    # mode (optional, default: single)
    regexMode = 0
    _regexMode = regexDef.get('mode', None)
    if _regexMode != None:
        if _regexMode == 'line':
            regexMode = 1
            cliCommandDef['regexes_byline'] += 1

    # re.match (optional, default: false)
    regexOp = re.search
    if regexDef.get('re.match', False):
        regexOp = re.match

    # regex
    regex = regexDef['regex']
    log.debug('adding: regex: %s, mode: %d, op: %s' % (regex, regexMode, regexOp))

    try:
        return (re.compile(regex), regex, regexMode, regexOp)
    except re.error as e:
        log.critical("Fatal: re compile failure: %s" % e)
        log.critical("cliCommandDef: %s" % cliCommandDef)
        log.critical("regex: %s, mode: %d, op: %s" % (regex, regexMode, regexOp))
        raise


def parseArCliTable(cliCommandDef, tableDef, fieldNameSeparator, mappings=None):
    cliCommandTable = dict()

    # Table name
    tableName = tableDef['name']
    cliCommandTable['name'] = tableName

    # marker (optional, default: None)
    cliCommandTable['marker'] = tableDef.get('marker', None)
    log.debug('table: %s marker: %s' % (tableName, cliCommandTable['marker']))

    # type (optional, default: table)
    cliCommandTable['type'] = 0
    tableKind = tableDef.get('type', None)
    if tableKind != None:
        if tableKind == 'kvp':
            cliCommandTable['type'] = 1
    log.debug('table: %s type: %s' % (tableName, cliCommandTable['type']))

    # mappings (optional, default: None)
    tableMappings = tableDef.get('mappings', None)
    if tableMappings != None:
        for (k, v) in tableMappings.items():
            # Table fields are of the form: <name>_<column>
            _updateMapping(mappings, cliCommandDef['document'], tableName + fieldNameSeparator + k, v)
    log.debug('table: %s mappings: %s' % (tableName, tableMappings))

    # timestamp (optional, default: None)
    cliCommandTable['timestamp'] = None
    cliCommandTable['idkey'] = None
    tableTimestamp = tableDef.get('timestamp', None)
    if tableTimestamp != None:
        cliCommandTable['timestamp'] = compile(tableTimestamp, tableTimestamp, 'eval')
        idkey = tableDef['idkey']
        cliCommandTable['idkey'] = compile(idkey, idkey, 'eval')
        log.debug('table: %s timestamp: %s idkey: %s' % (tableName, cliCommandTable['timestamp'],
                                                         cliCommandTable['idkey']))

    # regexes (optional, default: None)
    cliCommandTable['regexes'] = []
    tableRegexes = tableDef.get('regexes', None)
    if tableRegexes != None:
        for tableRegex in tableRegexes:
            cliCommandTableRegex = dict()
            cliCommandTableRegex['column'] = tableRegex['column']
            regexOp = re.search
            if tableRegex.get('re.match', False):
                regexOp = re.match
            cliCommandTableRegex['regex'] = \
                (re.compile(tableRegex['regex']), tableRegex['regex'], regexOp)
            regexMappings = tableRegex.get('mappings', None)
            if regexMappings != None:
                for (k, v) in regexMappings.items():
                    # Table fields are of the form: <name>_<column>
                    _updateMapping(mappings, cliCommandDef['document'], tableName + fieldNameSeparator + k, v)
            cliCommandTable['regexes'].append(cliCommandTableRegex)
            cliCommandTableRegex['when'] = tableRegex.get('when', None)
            cliCommandTableRegex['defaults'] = tableRegex.get('defaults', None)

    # begin and end (optional, default: None)
    cliCommandTable['begin'] = (None, None)
    tableBegin = tableDef.get('begin', None)
    if tableBegin != None:
        cliCommandTable['begin'] = (re.compile(tableBegin, flags=re.MULTILINE), tableBegin)
    cliCommandTable['end'] = (None, None)
    tableEnd = tableDef.get('end', None)
    if tableEnd != None:
        cliCommandTable['end'] = (re.compile(tableEnd, flags=re.MULTILINE), tableEnd)

    # table options (optional, default: None)
    cliCommandTable['toptions'] = None
    tableOptions = tableDef.get('toptions', None)
    if tableOptions != None:
        cliCommandTable['toptions'] = tableOptions
        if tableOptions.get('parser', None) is None:
            # Add default parser: aos-std
            cliCommandTable['toptions']['parser'] = 'aos-std'
        log.debug('table: %s toptions: %s' % (tableName, cliCommandTable['toptions']))

    return cliCommandTable


def parseArCliField(cliCommandDef, fieldDef, mappings):
    cliCommandField = dict()

    # Field name
    cliCommandField['ename'] = None
    if 'name' in fieldDef:
        # Simple field
        fieldName = fieldDef['name']
        cliCommandField['name'] = fieldName
    else:
        # Eval field
        fieldName = fieldDef['ename']
        cliCommandField['name'] = fieldName
        e = compile(fieldName, fieldName, 'eval')
        cliCommandField['ename'] = e
        log.debug('adding: ename: %s: %s' % (fieldName, e))

    # Field group can be overriden
    fieldGroup = fieldDef.get('group', cliCommandDef['group'])
    cliCommandField['group'] = fieldGroup

    # Parameter assignment
    fieldP = fieldDef.get('P', None)
    if fieldP != None:
        cliCommandField['P'] = fieldP
        log.debug('adding: %s, P: %s' % (fieldName, fieldP))

    # Column value assignment
    fieldC = fieldDef.get('C', None)
    if fieldC != None:
        cliCommandField['C'] = fieldC
        log.debug('adding: %s, R: %s' % (fieldName, fieldC))

    # eval() assignment
    fieldEval = fieldDef.get('eval', None)
    if fieldEval != None:
        e = compile(fieldEval, fieldEval, 'eval')
        cliCommandField['eval'] = e
        log.debug('adding: %s, eval: %s' % (fieldName, e))

    # when condition
    cliCommandField['when'] = None
    fieldWhen = fieldDef.get('when', None)
    if fieldWhen != None:
        e = compile(fieldWhen, fieldWhen, 'eval')
        cliCommandField['when'] = e
        log.debug('adding: %s, when: %s' % (fieldName, e))

    # mapping (optional, default: None)
    fieldMapping = fieldDef.get('mapping', None)
    if fieldMapping != None:
        if cliCommandField['ename'] != None:
            # ename fields are not supported yet!
            log.error('ignoring mapping for ename: %s' % (fieldName))
        else:
            _updateMapping(mappings, cliCommandDef['document'], fieldName, fieldMapping)
            log.debug('adding: %s, mapping: %s' % (fieldName, fieldMapping))

    return cliCommandField


def _updateMapping(mappings, document, fieldName, fieldMapping):
    mappings[fieldName] = fieldMapping
    root = mappings.get('_', None)
    if root is None:
        root = dict()
        mappings['_'] = root
    fields = root.get(document, None)
    if fields is None:
        fields = dict()
        root[document] = fields
    _fieldMapping = fields.get(fieldName, None)
    if _fieldMapping is not None and _fieldMapping != fieldMapping:
        log.error("FATAL: conflicting mapping for: %s/%s/%s" % (fieldName, _fieldMapping, fieldMapping))
        exit(1)
    fields[fieldName] = fieldMapping





#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Implements a generic table CLI output parser.
#
# All offset are measured relative to the first non blank line in the input.
#
# Three use-cases are handled:
# - header driven split-up based on regex with named groups
# - header driven split-up based on spacing between column names
# - data driven split-up based on regex with named groups
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
import re
from .AbstractTableParser import AbstractTableParser


# Logger
log = logging.getLogger(__name__)


class GenericRegexTableParser(AbstractTableParser):
    def __init__(self, source, tableOptions={}):
        super().__init__()
        self.source = source
        self.headerOffset = tableOptions.get("header.offset", -1)
        self.headerRegex = tableOptions.get("header.regex", None)
        self.dataOffset = tableOptions.get("data.offset", -1)
        self.dataRegex = tableOptions.get("data.regex", None)
        self.stopOnEmptyLine = tableOptions.get("stop.on.empty.line", True)
        self.skipPartialMatch = tableOptions.get("skip.partial.match", False)

    def readContent(self):
        if type(self.source) == list:
            self.contentLines = self.source
        else:
            # Sometimes multiple spurious \r are seen before real \r\n
            self.contentLines = re.sub(r"\r+\n", "\n", self.source).splitlines()

    def parse(self):
        # Loop until find first non blank line => reference, i.e. offset 0
        currentLineNo = 0
        while currentLineNo < len(self.contentLines):
            if self.contentLines[currentLineNo].strip() != "":
                break
            currentLineNo += 1

        columns = None
        pattern = None

        if (self.headerOffset >= 0) and ((currentLineNo + self.headerOffset) < len(self.contentLines)):
            # Create named positional columns, data lines chopped based begin/end
            headerLine = self.contentLines[currentLineNo + self.headerOffset]

            if self.headerRegex is not None:
                # Create named positional columns, data lines chopped based begin/end
                columns = self.createPositionalColumnsFromPattern(self.headerRegex, headerLine)
            else:
                # Create positional columns, data lines chopped based begin/end
                columns = self.createPositionalColumnsFromSplit(" +", headerLine)

        else:
            # Create named columns, data lines chopped based on pattern
            pattern = re.compile(self.dataRegex)
            columns = self.createColumnsFromPattern(pattern)

        # Main table
        self.getTables()[0]['name'] = "generic regex table"
        rows = []
        self.tables[0]['rows'] = rows
        self.tables[0]['header'] = columns

        # Read data
        if self.dataOffset < 0:
            return

        currentLineNo += self.dataOffset
        while currentLineNo < len(self.contentLines):
            currentLine = self.contentLines[currentLineNo]
            if currentLine.strip() == "":
                if self.stopOnEmptyLine:
                    break
                else:
                    currentLineNo += 1
                    continue

            if self.headerOffset >= 0:
                # Header based positional break-up
                dataRow = []
                for column in columns:
                    value = ""
                    begin = column['begin']
                    if begin < 0:
                        # Should never happen
                        begin = 0
                    end = column.get('end', -1)
                    if end < 0:
                        # Unset end means until end of line
                        end = len(currentLine)
                    if begin < len(currentLine):
                        value = currentLine[begin:min(end, len(currentLine))]
                    dataRow.append(value)
                rows.append(dataRow)

            else:
                # Data pattern based break-up
                matcher = pattern.match(currentLine)
                matched = matcher is not None
                if not matched and self.skipPartialMatch:
                    currentLineNo += 1
                    continue

                dataRow = []
                groupCount = len(matcher.groups()) if matched else 0
                for column in columns:
                    groupId = column['#']
                    value = ""
                    if groupId >= 1 and groupId <= groupCount:
                        value = matcher.group(groupId)
                        if value is None:
                            value = ""
                    dataRow.append(value)
                rows.append(dataRow)

            currentLineNo += 1

    def createColumnsFromPattern(self, pattern):
        columns = []
        groupIndex = pattern.groupindex
        for namedGroup, groupIndex in sorted(groupIndex.items(), key=lambda x: x[1]):
            column = { 'title': namedGroup, '#': groupIndex }
            columns.append(column)
        return columns

    def createPositionalColumnsFromPattern(self, regex, line):
        pattern = re.compile(regex)
        groupNameByGroupIndex = { }
        for g, i in pattern.groupindex.items():
            groupNameByGroupIndex[i] = g
        matcher = pattern.match(line)
        matched = matcher is not None
        if matched:
            columns = []
            c = 1
            groups = matcher.groups()
            for _ in groups:
                namedGroup = groupNameByGroupIndex.get(c, None)
                if namedGroup is not None:
                    column = { 'title': namedGroup, 'begin': matcher.start(c) }
                    if c < len(groups):
                        # Not last column
                        column['end'] = matcher.end(c)
                    columns.append(column)
                c += 1
            return columns
        else:
            raise Exception("[" + line + "]: does not match regex: [" + regex + "]")

    def createPositionalColumnsFromSplit(self, splitRegex, line):
        splits = re.split(splitRegex, line)
        columns = []
        i = 0
        for split in splits:
            column = { 'title': split, 'begin': line.index(split) }
            if (i + 1) < len(splits):
                # Not last column
                column['end'] = line.index(splits[i + 1])
            columns.append(column)
            i += 1
        return columns



#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# ARCLI result.
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

'''
#
# ArCliResult: ARCLI parser result.
#
'''
class ArCliResult():
    # Result types
    TYPE_REGEXES_BYLINE='regexes_byline'
    TYPE_REGEXES_ENTIRE='regexes_entire'
    TYPE_TABLES='tables'

    def __init__(self):
        self._result = {}

    def get(self):
        return self._result

    def addType(self, rtype):
        self._result[rtype] = []

    def addSection(self, rtype, sname=None):
        messages = []
        if sname is None:
            sname = rtype
        resultData = { sname: messages }
        self._result[rtype].append(resultData)
        return messages



#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# ArCliParserRequest
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
from ..CliParserBase import CliParserRequest


# Logger
log = logging.getLogger(__name__)


'''
#
# ArCliParserRequest: ARCLI parser request.
#
# MANDATORY fields
# - Source: <string> <source>
# - Command: <string> <CLI command>
# - Stdout: <string> <CLI output to parse>
# - Label: <string> <<device class>/<qualifier>>
# - LocalBeginTime: <float> <timestamp>
#
# OPTIONAL fields
# - Status: <int> <when < 0 parser will not process request>
# - ContentType: <string> <text/plain, application/json, etc...>
# - Origin: <string or dict> <origin tags>
#
'''
class ArCliParserRequest(CliParserRequest):
    def __init__(self):
        super().__init__()

    # Hook to provide additional output tags
    # Return None or a dictionary
    def getOutputTags(self, arCliContext, source):
        log.debug("arCliContext: %s, source: %s" % (arCliContext, source))
        outputTags = arCliContext.getOriginTags(self.get('Origin', None), source)
        return outputTags

    # Hook to provide additional body tags
    # MUST return a dictionary
    def getBodyTags(self, arCliContext, source):
        log.debug("arCliContext: %s, source: %s" % (arCliContext, source))
        return dict()

    # Hook for document processing
    # MUST return True to keep processing or False to stop processing
    def processDocument(self, arCliContext, source, cliCommandDef):
        document = cliCommandDef['document']
        group = cliCommandDef['group']
        log.debug("arCliContext: %s, source: %s, document: %s, group: %s" % (arCliContext, source, document, group))
        return True

    # Hook to process posted message (i.e. modify body and output)
    # MUST return array of (document, output) to post message(s) or None to discard message
    def updateMessageTags(self, arCliContext, source, document, output):
        log.debug("arCliContext: %s, source: %s, document: %s, output: %s" % (arCliContext, source, document, output))
        return [(document, output)]



#
# Copyright 2021 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# ARCLI parser context.
#
# Authors: Thomas Bastian, Jeff Goff
#
'''

'''
# Parameters
'''
'''
#
'''

import base64
import logging
from .ArCliDocuments import ArCliDocuments
from json import loads


# Logger
log = logging.getLogger(__name__)


# Default field name separator
ND_FIELD_NAME_SEPARATOR='.'


'''
#
# ArCliContext: ARCLI parser context.
#
'''
class ArCliContext:
    def __init__(self):
        self.arCliDocuments = None
        self.options = None
        self.origin = None
        self.originTagCache = { }
        self.originTagCacheBySource = { }

    def getArCliDocumentsByLabel(self, label):
        # First lookup by exact label
        arCliDocuments = self.arCliDocuments.getByLabel(label)
        if arCliDocuments is None:
            # Fallback to <device class>/default
            _label = label.split("/")[0] + "/default"
            arCliDocuments = self.arCliDocuments.getByLabel(_label)
            if arCliDocuments is None:
                # Fallback to default/default
                arCliDocuments = self.arCliDocuments.getByLabel("default/default")
        return arCliDocuments

    def loadArCliDocuments(self, contentProvider):
        self.arCliDocuments = ArCliDocuments(contentProvider, ND_FIELD_NAME_SEPARATOR)

    def getArCliMappings(self):
        return self.arCliDocuments.getMappings()

    # Set options
    def setOptions(self, options):
        self.options = options

    # Get options
    def getOptions(self):
        return self.options

    def setOrigin(self, origin):
        '''Set origin value. Supports:
        1. A dict
        2. A string that contains a list of comma separated key-value pairs:
           key1=value1[,key2=value]
        3. A base64 encoded JSON string
        '''
        self.origin = origin
        self.originTagCache = {}
        self.originTagCacheBySource = {}

    def getOrigin(self):
        return self.origin

    def _getOriginTags(self, origin):
        # Retrieve cached tags for a given origin object
        # Supported origin objects:
        # - dict object
        # - base64 encoded JSON
        # - "key1=value1[,key2=value]"
        if origin is None:
            return None
        if isinstance(origin, dict):
            # Dict are not hashable, hence cannot cache
            return origin
        tags = self.originTagCache.get(origin, None)
        if tags is None:
            ck = origin
            if isinstance(origin, str) or isinstance(origin, unicode):
                # base64 encoded JSON or "key1=value1[,key2=value]"
                if (len(origin) % 4) == 0:
                    # Possibly base64 encoded JSON
                    try:
                        _origin = base64.b64decode(origin)
                        origin = loads(_origin)
                    except TypeError:
                        log.debug('base64 decode failed: %s' % (origin))
                    except Exception:
                        log.error('origin: [%s] not JSON' % (_origin))
            if isinstance(origin, dict):
                tags = {}
                for k in origin.keys():
                    tags[k] = origin[k]
                self.originTagCache[ck] = tags
            elif isinstance(origin, str) or isinstance(origin, unicode):
                tags = {}
                kvps = origin.split(',')
                for _kvp in kvps:
                    kvp = _kvp.split('=', 1)
                    tags[kvp[0]] = kvp[1]
                self.originTagCache[ck] = tags
            else:
                raise Exception('unsupported origin type: ' + str(type(origin)))
        return tags

    def getOriginTags(self, origin, source):
        # Retrieve cached tags for a given origin object and source
        # Supported origin objects:
        # - dict object
        # - base64 encoded JSON
        # - "key1=value1[,key2=value]"
        # Origin object can be either:
        # - flat: { "customerid": "customerid" }
        # - keyed by source: { "<source1>": { "customerid": "customerid" }, ... }
        if origin is None:
            return None
        if source is None:
            # Must resolve based on source
            raise Exception("source is None")
        tags = self.originTagCacheBySource.get(source, None)
        if tags is None:
            # Get tags from origin
            otags = self._getOriginTags(origin)
            tags = {}
            if source in otags:
                # Keyed by source
                otags = otags[source]
            elif 'default' in otags:
                otags = otags['default']
            for k in otags.keys():
                if k.startswith('nd.origin.'):
                    tags[k] = otags[k]
                else:
                    if not isinstance(otags[k], dict):
                        tags['nd.origin.' + k] = otags[k]
            self.originTagCacheBySource[source] = tags
        if len(tags) == 0:
            return None
        return tags



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# get_my_ip() returns the outgoing interface IP address of the local machine.
#
# Authors: Albert Pang
#
'''

'''
# Parameters
'''
'''
#
'''

import socket  # noqa: E402
try:
    # Python 3
    from urllib.request import urlopen
    IS_PYTHON3 = True
except ImportError:
    IS_PYTHON3 = False
    import urllib2

# An reachable external IP address to test out-going traffic
TEST_IP_ADDRESS = "8.8.8.8"

EXTERNAL_IP_CHECK_URLS = [
    "http://ipinfo.io/ip",
    "https://checkip.amazonaws.com",
    "http://myip.dnsomatic.com",
    "https://ident.me"
]


def get_my_ip():
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.connect((TEST_IP_ADDRESS, 80))
    return s.getsockname()[0]

# https://stackoverflow.com/questions/2311510/getting-a-machines-external-ip-address-with-python


def get_my_external_ip3(url):
    return urlopen(url).read().decode('utf8').strip()


def get_my_external_ip2(url):
    req = urllib2.Request(url, data=None)
    response = urllib2.urlopen(req, timeout=5)

    return response.read().strip()


def get_my_external_ip():
    external_ip = "unknown"
    for url in EXTERNAL_IP_CHECK_URLS:
        try:
            if IS_PYTHON3:
                external_ip = get_my_external_ip3(url)
            else:
                external_ip = get_my_external_ip2(url)
            return external_ip
        except Exception:
            pass

    return external_ip


def main():
    print(get_my_ip())
    print(get_my_external_ip())


if __name__ == '__main__':
    main()



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# AirRecorder utilities.
#
# Authors: Thomas Bastian
#
'''

import logging
import time


log = logging.getLogger(__name__)


class Query():
    '''Python variant of AirRecorder Query.
    '''
    def __init__(self, command):
        self.command = command
        # Optional list of sub-queries
        self.subQueries = None
        # Optional {tag, value} pairs
        self.tags = None

    def setCommand(self, command):
        self.command = command

    def getCommand(self):
        return self.command

    def addSubQuery(self, subQuery):
        if self.subQueries is None:
            self.subQueries = []
        self.subQueries.append(subQuery)

    def getSubQueries(self):
        return self.subQueries

    def setTags(self, tags):
        self.tags = tags

    def getTags(self):
        return self.tags

    def addTag(self, tag, value):
        if self.tags is None:
            self.tags = dict()
        self.tags[tag] = value


class SubQuery():
    '''Python variant of AirRecorder SubQuery.
    '''
    def __init__(self, prompt, command, timeout):
        self.prompt = prompt
        self.command = command
        self.timeout = timeout


class Result():
    '''Python variant of AirRecorder Result.
    '''
    def __init__(self, query):
        self.query = query
        self.status = -1
        self.localBeginTime = None
        self.localEndTime = None
        self.remoteBeginTime = None
        self.remoteEndTime = None
        self.remoteRawTime = None
        self.stdout = ""
        self.stderr = ""
        self.throwable = None
        # Optional {tag, value} pairs
        self.resultTags = None

    def setQuery(self, query):
        self.query = query

    def getQuery(self):
        return self.query

    def setStatus(self, status):
        self.status = status

    def getStatus(self):
        return self.status

    def setLocalBeginTime(self, localBeginTime):
        self.localBeginTime = localBeginTime

    def getLocalBeginTime(self):
        return self.localBeginTime

    def setLocalEndTime(self, localEndTime):
        self.localEndTime = localEndTime

    def getLocalEndTime(self):
        return self.localEndTime

    def setRemoteBeginTime(self, remoteBeginTime):
        self.remoteBeginTime = remoteBeginTime

    def getRemoteBeginTime(self):
        return self.remoteBeginTime

    def setRemoteEndTime(self, remoteEndTime):
        self.remoteEndTime = remoteEndTime

    def getRemoteEndTime(self):
        return self.remoteEndTime

    def setRemoteRawTime(self, remoteRawTime):
        self.remoteRawTime = remoteRawTime

    def getRemoteRawTime(self):
        return self.remoteRawTime

    def setStdout(self, stdout):
        self.stdout = stdout

    def getStdout(self):
        return self.stdout

    def setStderr(self, stderr):
        self.stderr = stderr

    def getStderr(self):
        return self.stderr

    def getThrowable(self):
        return self.throwable

    def setResultTags(self, tags):
        self.resultTags = tags

    def getResultTags(self):
        return self.resultTags

    def addResultTag(self, tag, value):
        if self.resultTags is None:
            self.resultTags = dict()
        self.resultTags[tag] = value


'''
#
# Parse an AirRecorder log file and yield on each result

/////
///// Message: RESULT
///// Status: 1319
///// LocalBeginTime: 1311238912292 (2011-07-21T11:01:52.292+0200)
///// LocalEndTime: 1311238912326 (2011-07-21T11:01:52.326+0200)
///// RemoteBeginTime: 1311238824000 (2011-07-21T11:00:24.000+0200)
///// RemoteEndTime: 1311238824034 (2011-07-21T11:00:24.034+0200)
///// Command: show ap debug client-table ap-name 00:24:6c:ce:69:92
///// Section: Stdout
[text]
[///// Section: xxxxx
[text]
.....

#
'''


def _endSection(section, result, lines):
    if (section is not None):
        if (section.startswith('Stdout')):
            result.setStdout(lines)


def parseLogFileGenerator(fileName, tailSleepTime, sleep=time.sleep):
    messageBegin = "///// \n"
    messageEnd = ".....\n"
    result = None
    section = None
    lines = ""
    # AirRecorder header lines are always terminated with \r\n
    # Python3 universal newlines are disabled to ensure that the section content (i.e. Stdout/Stderr)
    # is read in as-is.
    logFile = open(fileName, 'r', newline='\n')
    while True:
        line = logFile.readline()
        if not line:
            # EOF reached
            if tailSleepTime < 0:
                logFile.close()
                return
            if tailSleepTime > 0:
                sleep(tailSleepTime)
            continue

        line_ = line.replace("\r", "")
        if line_ == messageBegin:
            # Message begin
            lines = ""
            query = Query(None)
            result = Result(query)
            continue

        if line_.startswith("///// "):
            # Message header
            _endSection(section, result, lines)
            line = line.rstrip()
            header = line[6:].split(' ')
            tag = header[0]
            if (tag.startswith('Message:')):
                pass
            elif (tag.startswith('Status:')):
                result.setStatus(int(header[1]))
            elif (tag.startswith('LocalBeginTime:')):
                result.setLocalBeginTime(int(header[1]))
            elif (tag.startswith('LocalEndTime:')):
                result.setLocalEndTime(int(header[1]))
            elif (tag.startswith('RemoteBeginTime:')):
                result.setRemoteBeginTime(int(header[1]))
            elif (tag.startswith('RemoteEndTime:')):
                result.setRemoteEndTime(int(header[1]))
            elif (tag.startswith('RemoteRawTime:')):
                # 6(///// ) + 14(RemoteRawTime:) + 1( )
                result.setRemoteRawTime(line[21:])
            elif (tag.startswith('QueryTag:')):
                # 6(///// ) + 9(QueryTag:) + 1( )
                queryTagRaw = line[16:].split('=', 2)
                result.getQuery().addTag(queryTagRaw[0], queryTagRaw[1])
            elif (tag.startswith('ResultTag:')):
                # 6(///// ) + 10(ResultTag:) + 1( )
                resultTagRaw = line[17:].split('=', 2)
                result.addResultTag(resultTagRaw[0], resultTagRaw[1])
            elif (tag.startswith('Command:')):
                # 6(///// ) + 8(Command:) + 1( )
                command = line[15:]
                result.getQuery().setCommand(command)
            elif (tag.startswith('Section:')):
                section = header[1]
            continue

        if line_ == messageEnd:
            # Message ends
            _endSection(section, result, lines)
            yield result
            result = None
            continue

        # Otherwise consume text
        if section is not None:
            lines += line


class LogFileReaderGenerator():
    '''Parse an AirRecorder log file and generate all results
    '''
    def __init__(self, fileName, tailSleepTime=-1):
        self._fileName = fileName
        self._tailSleepTime = tailSleepTime

    def __iter__(self):
        self._g = parseLogFileGenerator(self._fileName, self._tailSleepTime)
        return self

    # Python 3 compatibility
    def __next__(self):
        return next(self._g)


class LogFileProcessor():
    '''Parse an AirRecorder log file and call processResult() for each RESULT
    '''
    def __init__(self, fileName, tailSleepTime=-1):
        self._fileName = fileName
        self._tailSleepTime = tailSleepTime

    def run(self):
        # Generate results
        results = LogFileReaderGenerator(self._fileName, tailSleepTime=self._tailSleepTime)
        for result in results:
            if result.getStatus() <= 0:
                continue
            query = result.getQuery()
            if query is None:
                continue
            command = query.getCommand()
            if command is None:
                continue
            self.processResult(command, result)

    def processResult(self, command, result):
        pass



from logging import getLogger

from munch import DefaultMunch

import jstyleson
from yaml import load, FullLoader
from yaml.parser import ParserError


log = getLogger(__name__)


def convertToDict(s: str):
    ''' Attempt to decode a string and returns a dictionary or a list.
    Will attempt to decode the string as JSON or YAML

    Parameter
    ---------
    s : str

    Returns
    -------
    dict or list :
        A dictionary if able to convert content as dict or list. None if unsuccessful
    '''

    try:
        d = jstyleson.loads(s)
        log.debug("Found JSON")
        return DefaultMunch(None).fromDict(d)
    except ValueError as e:
        try:
            log.debug(f"Not JSON format ({e})")
            d = load(s, Loader=FullLoader)
            log.debug("Found YAML")
            return DefaultMunch(None).fromDict(d)
        except ParserError:
            log.debug("Did not find JSON or YAML str")
            return None



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Standard AOS row/column table parser.
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
import re
from .AbstractTableParser import AbstractTableParser


# Logger
log = logging.getLogger(__name__)


class RowColumnTableParserAosStd(AbstractTableParser):
    def __init__(self, source, marker=None):
        super().__init__()
        self.source = source
        self.marker = marker

    def readContent(self):
        if type(self.source) == list:
            self.contentLines = self.source
        else:
            self.contentLines = self.source.splitlines()

    def parse(self):
        # Process structure
        log.debug("marker [%s]" % str(self.marker))
        if self.marker is None:
            # No marker was given, find out from first 10 lines
            currentLineNo = 0
            for currentLine in self.contentLines:
                if currentLineNo > 10:
                    break
                if len(currentLine) > 0 and ((currentLineNo + 1) < len(self.contentLines)):
                    nextLine = self.contentLines[currentLineNo + 1]
                    if (re.match("\\-+", nextLine) is not None) and (len(currentLine) == len(nextLine)):
                        # Got marker
                        self.marker = currentLine
                        log.debug("got marker [%s]" % str(self.marker))
                        break
                currentLineNo += 1
            if self.marker is None:
                raise Exception()

        # Update main table
        self.tables[0]['name'] = self.marker
        log.debug("processing table [%s]" % str(self.marker))

        markerLine = -1
        currentPage = None
        currentLineNo = 0
        rows = []
        for currentLine in self.contentLines:
            # Begin page
            if currentLine.startswith(self.marker):
                log.debug("begin page [%s] at [%d]" % (str(self.marker), currentLineNo))
                markerLine = currentLineNo
                currentPage = None
                currentLineNo += 1
                continue

            # End page (TODO: maybe make empty line processing this an option?)
            if (len(currentLine.strip()) == 0) and (currentPage is not None):
                log.debug("end page [%s] at [%d]" % (str(self.marker), currentLineNo))
                markerLine = -1
                currentPage = None
                currentLineNo += 1
                continue

            # Skip information content between begin page and actual data
            if (currentPage is None) and (markerLine >= 0) and (currentLineNo > (markerLine + 2)):
                if currentLine.startswith("-"):
                    currentPage = self.createPage(self.tables[0],
                                                  self.contentLines[currentLineNo],
                                                  self.contentLines[currentLineNo - 1])
                    log.debug("new page [%s] at [%d]" % (currentPage, currentLineNo))
                    if not 'header' in self.tables[0]:
                        self.tables[0]['header'] = currentPage
                    currentLineNo += 1
                    continue

            if (currentPage is not None) and (currentLineNo > (markerLine + 3)):
                row = self.createRow(currentLine, currentPage)
                log.debug("new row [%s] at [%d]" % (row, currentLineNo))
                rows.append(row)

            currentLineNo += 1

        self.tables[0]['rows'] = rows



# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


with open('README.rst', 'r', encoding='utf-8') as f:
    readme = f.read()

setup(
    name='aacommons',
    version='0.1.15',
    description='aacommons',
    long_description=readme,
    long_description_content_type='text/x-rst',
    author='Thomas Bastian, Jeffrey Goff, Albert Pang',
    url='',
    test_suite="tests",
    packages=find_packages(exclude=('tests', 'docs')),
    install_requires=[
        "jstyleson",
        "munch",
        "PyYAML",
        "redis",
        "requests",
        "urllib3"
    ]
)



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Row/column table parser.
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging
import re
from .AbstractTableParser import AbstractTableParser
from .GenericJsonCsvTableParser import GenericJsonCsvTableParser
from .GenericRegexTableParser import GenericRegexTableParser
from .RowColumnTableParserAosDpStd import RowColumnTableParserAosDpStd
from .RowColumnTableParserAosStd import RowColumnTableParserAosStd


# Logger
log = logging.getLogger(__name__)


class RowColumnTableParser(AbstractTableParser):
    def __init__(self, source, marker=None, tableOptions=None):
        super().__init__()

        # Pre-process source
        if (tableOptions is not None) and 'begin' in tableOptions:
            tableBegin = tableOptions['begin']
            m_start = re.search(tableBegin, source)
            if m_start is not None:
                source = source[m_start.start():]
                log.debug("table.begin regex: [%s] matched at: [%s]" % (tableBegin, m_start.start()))
            else:
                log.debug("table.begin regex: [%s] did NOT match" % (tableBegin))
        if (tableOptions is not None) and 'end' in tableOptions:
            tableEnd = tableOptions['end']
            m_end = re.search(tableEnd, source)
            if m_end is not None:
                source = source[0:m_end.start()]
                log.debug("table.end regex: [%s] matched at: [%s]" % (tableEnd, m_end.start()))
            else:
                log.debug("table.end regex: [%s] did NOT match" % (tableEnd))
        if log.isEnabledFor(logging.DEBUG):
            # Log beginning and end of source
            b = min(10, len(source))
            e = min(10, len(source))
            log.debug("source.begin: [%s]" % (source[0:b]))
            log.debug("source.end: [%s]" % (source[len(source) - e:]))

        # Find delegate parser
        parserName = "aos-std"
        if tableOptions is not None:
            parserName = tableOptions['parser']
        if parserName == "aos-std":
            # Backwards compatibility, default to standard AOS tables
            self.parser = RowColumnTableParserAosStd(source, marker)

        elif parserName == "aos-dp-std":
            self.parser = RowColumnTableParserAosDpStd(source, marker)

        elif parserName == "generic-regex":
            self.parser = GenericRegexTableParser(source, tableOptions)

        elif parserName == "generic-json-csv":
            self.parser = GenericJsonCsvTableParser(source, tableOptions)

        else:
            raise Exception("unknown table parser: " + parserName)

    def process(self):
        self.parser.process()

    def getTables(self):
        return self.parser.getTables()



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# RedisStore
#
# A python class for managing the data stored in the redis database.
#
# Authors: Albert Pang
#
'''
# pylint: disable=invalid-name

import json
import jstyleson
import logging
from redis import StrictRedis
from redis.exceptions import ConnectionError, TimeoutError

# Logger
log = logging.getLogger(__name__)


DEFAULT_REDIS_HOST = 'localhost'
DEFAULT_REDIS_PORT = 6379
DEFAULT_REDIS_DB = 0
DEFAULT_REDIS_TIMEOUT = 3  # Normal connection should be very very fast.


class RedisStore(object):
    """A class for managing data in a redis server"""
    def __init__(self, host=DEFAULT_REDIS_HOST, port=DEFAULT_REDIS_PORT,
                 db=DEFAULT_REDIS_DB, url=None, log=None):
        '''
        Parameters
        ----------
        host : str
            Redis host
        port : int
            Redis port
        db : int
            Redis DB number (typically = 0)
        url : str
            Supported format:
            redis://[[username]:[password]]@localhost:6379/0
            rediss://[[username]:[password]]@localhost:6379/0
            unix://[[username]:[password]]@/path/to/socket.sock?db=0

            if url is specified, the other 3 parameters will be ignored
        '''
        super(RedisStore, self).__init__()
        self._log = log
        if url:
            self._store = StrictRedis().from_url(url, socket_connect_timeout=DEFAULT_REDIS_TIMEOUT)
        else:
            self._store = StrictRedis(host=host, port=port, db=db,
                                      socket_connect_timeout=DEFAULT_REDIS_TIMEOUT)
        try:
            self.ping()
        except (ConnectionError, TimeoutError) as e:
            if url:
                url_str = url
            else:
                url_str = "redis://{}:{}/{}. {}".format(host, port, db, e)
            self._log.critical("Cannot connect to Redis server at: %s. Error: %s" % (url_str, e))

    def set(self, key, value, ex=None, px=None, nx=False, xx=False):
        '''Set the value at key ``name`` to ``value`
        ex: sets an expire flag on key ``name`` for ``ex`` seconds.
        px: sets an expire flag on key ``name`` for ``px`` milliseconds.

        nx: if set to True, set the value at key ``name`` to ``value`` only
          if it does not exist. (i.e. noclobber)

        xx: if set to True, set the value at key ``name`` to ``value`` only
          if it already exists.
        '''
        return self._store.set(key, value, ex=ex, px=px, nx=nx, xx=xx)

    def get(self, name):
        '''Return the value at key ``name``, or "" if the key doesn't exist'''
        value = self._store.get(name)
        if value:
            return value.decode('utf-8')
        else:
            return ""

    def delete(self, name):
        '''delete a key'''
        return self._store.delete(name)

    def exists(self, key):
        '''return True if key exist. False otherwise'''
        return self._store.exists(key)

    def ping(self):
        '''ping the server'''
        return self._store.ping()

    def store_file(self, key, filename, ex=None, px=None, nx=False, xx=False):
        '''stores the entire un-processed content of filename as value into key'''
        try:
            filecontent = open(filename, 'r').read()
        except (FileNotFoundError, PermissionError) as e:
            self._log.error("Cannot open %s for reading. %s" % (filename, e))
            return None
        return self.set(key, filecontent, ex=ex, px=px, nx=nx, xx=xx)

    def store_json_file(self, key, json_file, ex=None, px=None, nx=False, xx=False):
        '''stores the content of json_file as volue into key.
        Attempt to decode the content as JSON before store.
        '''
        try:
            filecontent = open(json_file, 'r').read()
        except (FileNotFoundError, PermissionError) as e:
            self._log.error("Cannot open %s for reading. %s" % (json_file, e))
            return None
        try:
            data = jstyleson.loads(filecontent)
        except ValueError as e:
            self._log.error("Cannot load %s as json object. File not in json format (%s)" % (json_file, e))
            return None
        return self.set(key, json.dumps(data))

    def keys(self, pattern='*'):
        '''Return all keys of the redis server'''
        return self._store.keys(pattern)

    @property
    def dbsize(self):
        '''returns the number of keys in db'''
        return self._store.dbsize()


if __name__ == '__main__':
    pass



#
# Copyright 2020 Thomas Bastian, Jeffrey Goff, Albert Pang
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

'''
#
# Abstract table parser.
#
# Authors: Thomas Bastian
#
'''

'''
# Parameters
'''
'''
#
'''

import logging


# Logger
log = logging.getLogger(__name__)


class AbstractTableParser():
    def __init__(self):
        self.tables = [{}]
        self.contentLines = []

    def process(self):
        self.preProcess()
        self.readContent()
        self.parse()
        self.postProcess()

    def preProcess(self):
        pass

    def readContent(self):
        pass

    def parse(self):
        pass

    def postProcess(self):
        pass

    def getTables(self):
        return self.tables

    def createPage(self, table, line, header):
        log.debug("table: [%s], line: [%s], header: [%s]" % (table, line, header))
        page = []
        begin = 0
        i = 0
        while i < len(line):
            while (i < len(line)) and (line[i] == '-'):
                i += 1
                continue
            while (i < len(line)) and (line[i] == ' '):
                i += 1
                continue
            column = { 'begin': begin, 'end': i, 'title': header[begin:i] }
            page.append(column)
            begin = i
        return page

    def createRow(self, current, currentPage):
        row = []
        columnNo = 1
        columnCount = len(currentPage)
        for headerColumn in currentPage:
            try:
                beginIndex = headerColumn['begin']
                endIndex = headerColumn['end']
                clen = len(current)
                if clen <= beginIndex:
                    # Not enough characters, generate empty string
                    endIndex = beginIndex = 0
                elif clen <= endIndex:
                    # Not enough characters for this column
                    endIndex = clen
                else:
                    if columnNo == columnCount:
                        # Last column eats everything up
                        endIndex = clen
                row.append(current[beginIndex:endIndex])
            except Exception as e:
                raise Exception(e)
            columnNo += 1
        return row
